import inspect
from functools import wraps
import os
from types import UnionType
from kisa_utils.response import Response, Ok, Error
from kisa_utils.storage import Path
from kisa_utils.structures.validator import Value, validateWithResponse
import copy
from typing import Any

class Definition:
    '''
    this class is meant to be used in defininig arg and kwarg types without throwing funny errors from pylance
    usage:
        `Definition[definition]` -> equevalent to just `definition` as `Definition` is just a mirror
    eg

    def func(
        a:int, 
        b:Definition[{
            "key1": type,
            ...
        }],
        /
    ) -> retType:
        ...

    '''

    @staticmethod
    def __class_getitem__(definition:Any) -> Any:
        return definition


def private(func):
    '''
    function is used only in the same module as that in which its defined
    * if multiple decorators are used, this has to be the closest to the function being decorated
    '''
    @wraps(func)
    def wrapper(*args, **kwargs):
        expectedModule = inspect.getfile(func)
        stack = inspect.stack()
        stackCallerModules = [(s.function,s.filename) for s in stack if ('<frozen' not in s.filename)]

        callingModule = stackCallerModules[1][1]
        # print(Path.getMyAbsolutePath())
        # from pprint import pprint; pprint(stackCallerModules)
        # print(expectedModule, callingModule, expectedModule == callingModule)
        if expectedModule != callingModule:
            raise Exception(f"can't call private function `{func.__name__}` from external module")

        return func(*args, **kwargs)
    return wrapper

def protected(func):
    '''
    function is only allowed to be used in;
    * the module in which its defined
    * submodules starting in the defining module's path

    NB:
    * if multiple decorators are used, this has to be the closest to the function being decorated
    '''
    @wraps(func)
    def wrapper(*args, **kwargs):
        expectedModule = inspect.getfile(func)
        stack = inspect.stack()
        stackCallerModules = [(s.function,s.filename) for s in stack if ('<frozen' not in s.filename)]

        callingModule = stackCallerModules[1][1]
        # print(expectedModule, callingModule)
        if not callingModule.startswith(os.path.dirname(expectedModule)):
            raise Exception(f"can't call protected function `{func.__name__}` from path outside definition module path")

        return func(*args, **kwargs)
    return wrapper

def returnsKISAResponse(func):
    '''
    this decorator should be applied above `private` and/or `protected`
    '''
    @wraps(func)
    def wrapper(*args, **kwargs) -> Response:
        if not isinstance(value := func(*args, **kwargs), Response):
            raise Exception(f'function `{func.__name__}` returned a non-KISA response')

        return value

    return wrapper

def returns(returnType):
    '''
    this decorator should be applied above `private` and/or `protected`
    '''
    if isinstance(returnType, UnionType):
        raise Exception(f'decorator `returns` does not accept union types as an argument')
    if type(returnType) != type(int):
        raise Exception(f'decorator `returns` only accepts a Type as an argument')
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if not isinstance(value:=func(*args, **kwargs), returnType):
                raise Exception(f'function `{func.__name__}` expected to return `{returnType}`, got `{type(value)}` instead')
            return value
        return wrapper
    return decorator

def enforceReturnType(func):
    '''
    * this decorator should be applied above `private` and/or `protected`
    '''
    returnType = func.__annotations__.get('return',None)

    if returnType == None:
        raise Exception(f'`{func.__name__}` has no return type indicated')

    if isinstance(returnType, UnionType):
        raise Exception(f'decorator `returns` does not accept union types as an argument')

    @wraps(func)
    def wrapper(*args, **kwargs):
        if not isinstance(value:=func(*args, **kwargs), returnType):
            raise Exception(f'function `{func.__name__}` expected to return `{returnType}`, got `{type(value)}` instead')
        return value
    return wrapper

def enforceTypes(func):
    if 1:
        signature = inspect.signature(func)
        typeHints = func.__annotations__
        parameters = signature.parameters

        if not func.__doc__:
            raise ValueError(
                f"function `{func.__name__}` has no docString!")

        if not typeHints.get("return", None):
            raise TypeError(
                f"function `{func.__name__}` has no return type")

        registeredArgs = []
        registeredKwargs = {}

        for key, value in signature.parameters.items():
            if value.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                raise TypeError(
                    f"function `{func.__name__}` should take either positional-only or keyword-only parameters. `{key}` is neither")

            if value.kind == inspect.Parameter.POSITIONAL_ONLY:
                if value.default is not inspect.Parameter.empty:
                    if not (resp := validateWithResponse(value.default, typeHints[key])):
                        raise ValueError(f'arg `{key}` default value: {resp.log}')
                registeredArgs.append((key, typeHints[key]))
            elif value.kind == inspect.Parameter.KEYWORD_ONLY:
                if value.default is not inspect.Parameter.empty:
                    if not (resp := validateWithResponse(value.default, typeHints[key])):
                        raise ValueError(f'kwarg `{key}` default value: {resp.log}')
                registeredKwargs[key] = typeHints[key]

            if str(value).startswith('*'):
                raise ValueError(
                    f'function `{func.__name__}`: *{key} or **{key} not allowed in function signature')

            hint = typeHints.get(key, None)
            if not hint:
                raise TypeError(f"parameter `{key}` has no type hint")

        expectectedReturnType = typeHints['return']

    @wraps(func)
    def w(*args, **kwargs):
        for index, arg in enumerate(args):
            if not (resp := validateWithResponse(arg, registeredArgs[index][1])):
                log = f'arg `{registeredArgs[index][0]}`, arg-index {index}: {resp.log}'
                if Response == expectectedReturnType: return Error(log)
                raise ValueError(log)

        for kwarg in kwargs:
            if kwarg not in registeredKwargs:
                print(f'[VALIDATOR WARNING]: kwarg `{kwarg}` was not part of function `{func.__name__}` definition so its type hints cant be enforced')
                continue

            if not (resp := validateWithResponse(kwargs[kwarg], registeredKwargs[kwarg])):
                log = f'kwarg `{kwarg}`: {resp.log}'
                if Response == expectectedReturnType: return Error(log)
                raise ValueError(log)

        resp = func(*args, **kwargs)

        if not isinstance(resp, expectectedReturnType):
            if isinstance(resp, type):
                log = f'`{func.__name__}` returned `{resp.__name__}`, expected `{expectectedReturnType}`'
            else:
                log = f'`{func.__name__}` returned `{type(resp).__name__}`, expected `{expectectedReturnType}`'
            
            if Response == expectectedReturnType: return Error(log)
            raise TypeError(log)

        return resp

    return w

